from __future__ import annotations

from hypothesis import given
from hypothesis.strategies import integers

from utilities.holoviews import plot_curve
from utilities.hypothesis import (
    assume_does_not_raise,
    float_data_arrays,
    int_indexes,
    text_ascii,
)
from utilities.xarray import DataArrayF1


class TestPlotCurve:
    @given(
        array=float_data_arrays(dim=int_indexes(), name=text_ascii(min_size=1)),
        smooth=integers(1, 10),
    )
    def test_smooth(self, *, array: DataArrayF1, smooth: int) -> None:
        with assume_does_not_raise(RuntimeWarning):
            curve = plot_curve(array, smooth=smooth)
        assert curve.label == f"{array.name} (MA{smooth})"



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from hypothesis.extra.numpy import array_shapes
from hypothesis.strategies import DrawFn, SearchStrategy, composite
from xarray import DataArray

from utilities._hypothesis.common import (
    MaybeSearchStrategy,
    lift_draw,
    lists_fixed_length,
    text_ascii,
)
from utilities._hypothesis.numpy import (
    bool_arrays,
    float_arrays,
    int_arrays,
    str_arrays,
)
from utilities._hypothesis.pandas import int_indexes
from utilities.pandas import IndexA, IndexI
from utilities.xarray import DataArrayB, DataArrayF, DataArrayI, DataArrayO


@composite
def bool_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayB:
    """Strategy for generating data arrays of booleans."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(bool_arrays(shape=shape, fill=fill, unique=unique))
    return DataArray(data=values, coords=indexes_, dims=list(indexes_), name=draw(name))


@composite
def dicts_of_indexes(
    _draw: DrawFn,
    /,
    *,
    min_dims: int = 1,
    max_dims: int | None = None,
    min_side: int = 1,
    max_side: int | None = None,
) -> dict[str, IndexI]:
    """Strategy for generating dictionaries of indexes."""
    draw = lift_draw(_draw)
    shape = draw(
        array_shapes(
            min_dims=min_dims, max_dims=max_dims, min_side=min_side, max_side=max_side
        )
    )
    ndims = len(shape)
    dims = draw(lists_fixed_length(text_ascii(), ndims, unique=True))
    indexes = (draw(int_indexes(n=length)) for length in shape)
    return dict(zip(dims, indexes, strict=True))


@composite
def float_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_value: MaybeSearchStrategy[float | None] = None,
    max_value: MaybeSearchStrategy[float | None] = None,
    allow_nan: MaybeSearchStrategy[bool] = False,
    allow_inf: MaybeSearchStrategy[bool] = False,
    allow_pos_inf: MaybeSearchStrategy[bool] = False,
    allow_neg_inf: MaybeSearchStrategy[bool] = False,
    integral: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayF:
    """Strategy for generating data arrays of floats."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        float_arrays(
            shape=shape,
            min_value=min_value,
            max_value=max_value,
            allow_nan=allow_nan,
            allow_inf=allow_inf,
            allow_pos_inf=allow_pos_inf,
            allow_neg_inf=allow_neg_inf,
            integral=integral,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(data=values, coords=indexes_, dims=list(indexes_), name=draw(name))


@composite
def int_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_value: MaybeSearchStrategy[int | None] = None,
    max_value: MaybeSearchStrategy[int | None] = None,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayI:
    """Strategy for generating data arrays of ints."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        int_arrays(
            shape=shape,
            min_value=min_value,
            max_value=max_value,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(data=values, coords=indexes_, dims=list(indexes_), name=draw(name))


@composite
def str_data_arrays(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    allow_none: MaybeSearchStrategy[bool] = False,
    fill: SearchStrategy[Any] | None = None,
    unique: MaybeSearchStrategy[bool] = False,
    name: MaybeSearchStrategy[str | None] = None,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> DataArrayO:
    """Strategy for generating data arrays of strings."""
    draw = lift_draw(_draw)
    indexes_ = draw(_merge_into_dict_of_indexes(indexes, **indexes_kwargs))
    shape = tuple(map(len, indexes_.values()))
    values = draw(
        str_arrays(
            shape=shape,
            min_size=min_size,
            max_size=max_size,
            allow_none=allow_none,
            fill=fill,
            unique=unique,
        )
    )
    return DataArray(data=values, coords=indexes_, dims=list(indexes_), name=draw(name))


@composite
def _merge_into_dict_of_indexes(
    _draw: DrawFn,
    indexes: MaybeSearchStrategy[Mapping[str, IndexA]] | None = None,
    /,
    **indexes_kwargs: MaybeSearchStrategy[IndexA],
) -> dict[str, IndexA]:
    """Merge positional & kwargs of indexes into a dictionary."""
    draw = lift_draw(_draw)
    if (indexes is None) and (len(indexes_kwargs) == 0):
        return draw(dicts_of_indexes())
    indexes_out: dict[str, IndexA] = {}
    if indexes is not None:
        indexes_out |= dict(draw(indexes))
    indexes_out |= {k: draw(v) for k, v in indexes_kwargs.items()}
    return indexes_out


__all__ = [
    "bool_data_arrays",
    "dicts_of_indexes",
    "float_data_arrays",
    "int_data_arrays",
    "str_data_arrays",
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from utilities.pathvalidate import valid_path
from utilities.typed_settings import click_field


@dataclass(frozen=True)
class Config:
    """Settings for the `pypi_server` script."""

    pid_file: Path = click_field(
        default=valid_path("pidfile"), param_decls=("-pf", "--pidfile")
    )
    log_dir: Path = click_field(
        default=valid_path("logs"), param_decls=("-ld", "--log-dir")
    )
    state_path: Path = click_field(
        default=valid_path("luigi-state.pickle"), param_decls=("-sp", "--state-path")
    )
    port: int = click_field(default=1456, param_decls=("-po", "--port"))
    dry_run: bool = click_field(default=False, param_decls=("-dr", "--dry-run"))
    exist_ok: bool = click_field(default=False, param_decls=("-e", "--exist-ok"))



from __future__ import annotations

from dataclasses import dataclass

from utilities.beartype import beartype_if_dev


@beartype_if_dev
def identity(x: int, /) -> int:
    return x


@beartype_if_dev
@dataclass
class Example:
    x: int


if __name__ == "__main__":
    # these should pass:
    _ = identity(0.0)  # type: ignore
    _ = Example(0.0)  # type: ignore



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager

from pyinstrument.profiler import Profiler

from utilities.atomicwrites import writer
from utilities.datetime import get_now, local_timezone
from utilities.pathvalidate import valid_path, valid_path_cwd
from utilities.types import PathLike


@contextmanager
def profile(*, path: PathLike = valid_path_cwd()) -> Iterator[None]:
    """Profile the contents of a block."""
    with Profiler() as profiler:
        yield
    now = get_now(tz=local_timezone())
    filename = valid_path(path, f"profile__{now:%Y%m%dT%H%M%S}.html")
    with writer(filename) as temp, temp.open(mode="w") as fh:
        _ = fh.write(profiler.output_html())


__all__ = ["profile"]



from __future__ import annotations

import datetime as dt
from collections.abc import Hashable, Mapping, Sized
from collections.abc import Set as AbstractSet
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TypeGuard, TypeVar, overload

from typing_extensions import override

from utilities.more_itertools import always_iterable

Number = int | float
Duration = Number | dt.timedelta
SequenceStrs = list[str] | tuple[str, ...]
IterableStrs = SequenceStrs | AbstractSet[str] | Mapping[str, Any]
PathLike = Path | str


_T = TypeVar("_T")


@overload
def get_class(obj: type[_T], /) -> type[_T]:
    ...


@overload
def get_class(obj: _T, /) -> type[_T]:
    ...


def get_class(obj: _T | type[_T], /) -> type[_T]:
    """Get the class of an object, unless it is already a class."""
    return obj if isinstance(obj, type) else type(obj)


def get_class_name(obj: Any, /) -> str:
    """Get the name of the class of an object, unless it is already a class."""
    return get_class(obj).__name__


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")


@overload
def ensure_class(obj: Any, cls: type[_T], /) -> _T:
    ...


@overload
def ensure_class(obj: Any, cls: tuple[type[_T1], type[_T2]], /) -> _T1 | _T2:
    ...


@overload
def ensure_class(
    obj: Any, cls: tuple[type[_T1], type[_T2], type[_T3]], /
) -> _T1 | _T2 | _T3:
    ...


@overload
def ensure_class(
    obj: Any, cls: tuple[type[_T1], type[_T2], type[_T3], type[_T4]], /
) -> _T1 | _T2 | _T3 | _T4:
    ...


@overload
def ensure_class(
    obj: Any, cls: tuple[type[_T1], type[_T2], type[_T3], type[_T4], type[_T5]], /
) -> _T1 | _T2 | _T3 | _T4 | _T5:
    ...


def ensure_class(obj: Any, cls: type[_T] | tuple[type[_T], ...], /) -> _T:  # type: ignore
    """Ensure an object is of the required class."""
    if isinstance(obj, cls):
        return obj
    raise EnsureClassError(obj=obj, cls=cls)


@dataclass(frozen=True, kw_only=True, slots=True)
class EnsureClassError(Exception):
    obj: Any
    cls: type[Any] | tuple[type[Any], ...]

    @override
    def __str__(self) -> str:
        return "Object {} must be an instance of {}; got {!r}".format(
            self.obj,
            ", ".join(map(repr, map(get_class_name, always_iterable(self.cls)))),
            get_class_name(self.obj),
        )


def ensure_hashable(obj: Any, /) -> Hashable:
    """Ensure an object is hashable."""
    if is_hashable(obj):
        return obj
    raise EnsureHashableError(obj=obj)


@dataclass(frozen=True, kw_only=True, slots=True)
class EnsureHashableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return "Object {} must be hashable".format(self.obj)


def is_hashable(obj: Any, /) -> TypeGuard[Hashable]:
    """Check if an object is hashable."""
    try:
        _ = hash(obj)
    except TypeError:
        return False
    return True


def issubclass_except_bool_int(x: type[Any], y: type[Any], /) -> bool:
    """Checks for the subclass relation, except bool < int."""
    return issubclass(x, y) and not (issubclass(x, bool) and issubclass(int, y))


def is_sized_not_str(obj: Any, /) -> TypeGuard[Sized]:
    """Check if an object is sized, but not a string."""
    try:
        _ = len(obj)
    except TypeError:
        return False
    return not isinstance(obj, str)


__all__ = [
    "Duration",
    "EnsureClassError",
    "EnsureHashableError",
    "IterableStrs",
    "Number",
    "PathLike",
    "SequenceStrs",
    "ensure_class",
    "ensure_hashable",
    "get_class",
    "get_class_name",
    "is_hashable",
    "is_sized_not_str",
    "issubclass_except_bool_int",
]



from __future__ import annotations

from collections.abc import Iterator, Mapping
from functools import partial
from itertools import chain, repeat, starmap
from re import MULTILINE, escape, search
from subprocess import PIPE, CalledProcessError, check_output
from typing import Any

from utilities.errors import redirect_error
from utilities.more_itertools import OneError, one
from utilities.os import temp_environ
from utilities.pathvalidate import valid_path, valid_path_cwd
from utilities.types import IterableStrs, PathLike


def get_shell_output(
    cmd: str,
    /,
    *,
    cwd: PathLike = valid_path_cwd(),
    activate: PathLike | None = None,
    env: Mapping[str, str | None] | None = None,
) -> str:
    """Get the output of a shell call.

    Optionally, activate a virtual environment if necessary.
    """
    cwd = valid_path(cwd)
    if activate is not None:
        with redirect_error(OneError, GetShellOutputError(f"{cwd=}")):
            activate = one(cwd.rglob("activate"))
        cmd = f"source {activate}; {cmd}"  # pragma: os-ne-windows

    with temp_environ(env):
        return check_output(
            cmd,
            stderr=PIPE,
            shell=True,  # noqa: S602
            cwd=cwd,
            text=True,
        )


class GetShellOutputError(Exception):
    ...


def run_accept_address_in_use(args: IterableStrs, /, *, exist_ok: bool) -> None:
    """Run a command, accepting the 'address already in use' error."""
    try:  # pragma: no cover
        _ = check_output(list(args), stderr=PIPE, text=True)  # noqa: S603
    except CalledProcessError as error:  # pragma: no cover
        pattern = _address_already_in_use_pattern()
        try:
            from loguru import logger
        except ModuleNotFoundError:
            info = exception = print
        else:
            info = logger.info
            exception = logger.exception
        if exist_ok and search(pattern, error.stderr, flags=MULTILINE):
            info("Address already in use")
        else:
            exception("Address already in use")
            raise


def _address_already_in_use_pattern() -> str:
    """Get the 'address_already_in_use' pattern."""
    text = "OSError: [Errno 98] Address already in use"
    escaped = escape(text)
    return f"^{escaped}$"


def tabulate_called_process_error(error: CalledProcessError, /) -> str:
    """Tabulate the components of a CalledProcessError."""
    mapping = {  # pragma: os-ne-windows
        "cmd": error.cmd,
        "returncode": error.returncode,
        "stdout": error.stdout,
        "stderr": error.stderr,
    }
    max_key_len = max(map(len, mapping))  # pragma: os-ne-windows
    tabulate = partial(_tabulate, buffer=max_key_len + 1)  # pragma: os-ne-windows
    return "\n".join(starmap(tabulate, mapping.items()))  # pragma: os-ne-windows


def _tabulate(key: str, value: Any, /, *, buffer: int) -> str:
    template = f"{{:{buffer}}}{{}}"  # pragma: os-ne-windows

    def yield_lines() -> Iterator[str]:  # pragma: os-ne-windows
        keys = chain([key], repeat(buffer * " "))
        value_lines = str(value).splitlines()
        for k, v in zip(keys, value_lines, strict=False):
            yield template.format(k, v)

    return "\n".join(yield_lines())  # pragma: os-ne-windows


__all__ = [
    "GetShellOutputError",
    "get_shell_output",
    "run_accept_address_in_use",
    "tabulate_called_process_error",
]



from __future__ import annotations

import datetime as dt
from collections.abc import Iterable, Iterator
from functools import partial
from getpass import getuser
from itertools import islice
from pathlib import Path
from shutil import rmtree

from click import command
from loguru import logger
from typed_settings import find

from utilities.datetime import UTC, get_now
from utilities.loguru import setup_loguru
from utilities.scripts.clean_dir.classes import Config, Item
from utilities.typed_settings import click_options
from utilities.types import PathLike

_CONFIG = Config()


@command()
@click_options(Config, appname="cleandir", config_files=[find("config.toml")])
def main(config: Config, /) -> None:
    """CLI for the `clean_dir` script."""
    setup_loguru()  # pragma: os-ne-windows
    if config.dry_run:  # pragma: os-ne-windows
        for item in _yield_items(  # pragma: os-ne-windows
            paths=config.paths, days=config.days, chunk_size=config.chunk_size
        ):
            logger.debug("{path}", path=item.path)  # pragma: os-ne-windows
    else:
        _clean_dir(  # pragma: os-ne-windows
            paths=config.paths, days=config.days, chunk_size=config.chunk_size
        )


def _clean_dir(
    *,
    paths: Iterable[PathLike] = _CONFIG.paths,
    days: int = _CONFIG.days,
    chunk_size: int | None = _CONFIG.chunk_size,
) -> None:
    while True:  # pragma: os-ne-windows
        iterator = _yield_items(  # pragma: os-ne-windows
            paths=paths, days=days, chunk_size=chunk_size
        )
        if len(items := list(iterator)) >= 1:  # pragma: os-ne-windows
            for item in items:  # pragma: os-ne-windows
                item.clean()  # pragma: os-ne-windows
        else:
            return


def _yield_items(
    *,
    paths: Iterable[PathLike] = _CONFIG.paths,
    days: int = _CONFIG.days,
    chunk_size: int | None = _CONFIG.chunk_size,
) -> Iterator[Item]:
    it = _yield_inner(paths=paths, days=days)  # pragma: os-ne-windows
    if chunk_size is not None:  # pragma: os-ne-windows
        return islice(it, chunk_size)  # pragma: os-ne-windows
    return it  # pragma: os-ne-windows


def _yield_inner(
    *, paths: Iterable[PathLike] = _CONFIG.paths, days: int = _CONFIG.days
) -> Iterator[Item]:
    for path in map(Path, paths):  # pragma: os-ne-windows
        for p in path.rglob("*"):  # pragma: os-ne-windows
            yield from _yield_from_path(p, path, days=days)  # pragma: os-ne-windows


def _yield_from_path(
    p: Path, path: Path, /, *, days: int = _CONFIG.days
) -> Iterator[Item]:
    p, path = map(Path, [p, path])  # pragma: os-ne-windows
    if p.is_symlink():  # pragma: os-ne-windows
        yield from _yield_from_path(
            p.resolve(), path, days=days
        )  # pragma: os-ne-windows
    elif _is_owned_and_relative(p, path):  # pragma: no cover
        if (p.is_file() or p.is_socket()) and _is_old(p, days=days):
            yield Item(p, partial(_unlink_path, p))
        elif p.is_dir() and _is_empty(p):
            yield Item(p, partial(_unlink_dir, p))


def _is_owned_and_relative(p: Path, path: Path, /) -> bool:
    try:  # pragma: os-ne-windows
        return (p.owner() == getuser()) and p.is_relative_to(path)
    except FileNotFoundError:  # pragma: no cover
        return False


def _is_empty(path: Path, /) -> bool:
    return len(list(path.iterdir())) == 0  # pragma: os-ne-windows


def _is_old(path: Path, /, *, days: int = _CONFIG.days) -> bool:
    age = get_now(tz=UTC) - dt.datetime.fromtimestamp(  # pragma: os-ne-windows
        path.stat().st_mtime, tz=UTC
    )
    return age >= dt.timedelta(days=days)  # pragma: os-ne-windows


def _unlink_path(path: Path, /) -> None:
    logger.info("Removing file:      {path}", path=path)  # pragma: os-ne-windows
    path.unlink(missing_ok=True)  # pragma: os-ne-windows


def _unlink_dir(path: Path, /) -> None:
    logger.info("Removing directory: {path}", path=path)  # pragma: os-ne-windows
    rmtree(path, ignore_errors=True)  # pragma: os-ne-windows



from __future__ import annotations

from zarr import Array

from utilities._zarr.common import (
    GetIndexByNameError,
    GetSelIndexerError,
    IselIndexer,
    NDArrayWithIndexes,
    yield_array_with_indexes,
    yield_group_and_array,
)
from utilities.numpy import _ffill_non_nan_slices_helper, array_indexer


def ffill_non_nan_slices(
    array: Array, /, *, limit: int | None = None, axis: int = -1
) -> None:
    """Forward fill the slices in an array which contain non-nan values."""
    ndim = array.ndim
    arrays = (
        array.oindex[array_indexer(i, ndim, axis=axis)]
        for i in range(array.shape[axis])
    )
    for i, repl_i in _ffill_non_nan_slices_helper(arrays, limit=limit):
        array.oindex[array_indexer(i, ndim, axis=axis)] = repl_i


__all__ = [
    "GetIndexByNameError",
    "GetSelIndexerError",
    "IselIndexer",
    "NDArrayWithIndexes",
    "ffill_non_nan_slices",
    "yield_array_with_indexes",
    "yield_group_and_array",
]


try:
    from utilities._zarr.xarray import (
        DataArrayOnDisk,
        ToNDArray1Error,
        save_data_array_to_disk,
        to_ndarray1,
        yield_data_array_on_disk,
    )
except ModuleNotFoundError:  # pragma: no cover
    pass
else:
    __all__ += [
        "DataArrayOnDisk",
        "ToNDArray1Error",
        "save_data_array_to_disk",
        "to_ndarray1",
        "yield_data_array_on_disk",
    ]



from __future__ import annotations

from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject, data, dictionaries, integers, none
from numpy import arange, array, zeros
from numpy.testing import assert_equal
from pandas import Index, RangeIndex
from pandas.testing import assert_index_equal
from pytest import mark, param, raises
from xarray import DataArray
from xarray.testing import assert_identical

from utilities._zarr.xarray import (
    DataArrayOnDisk,
    ToNDArray1Error,
    save_data_array_to_disk,
    to_ndarray1,
    yield_data_array_on_disk,
)
from utilities.hypothesis import (
    float_arrays,
    int_arrays,
    int_indexes,
    temp_paths,
    text_ascii,
)
from utilities.numpy import NDArrayI1
from utilities.pathvalidate import valid_path
from utilities.types import get_class_name
from utilities.warnings import suppress_warnings
from utilities.xarray import DataArray1


class TestDataArrayOnDisk:
    @given(
        data=data(),
        coords=dictionaries(text_ascii(), integers() | int_indexes(), max_size=3),
        name=text_ascii() | none(),
        root=temp_paths(),
    )
    def test_main(
        self, data: DataObject, coords: Mapping[str, Any], name: str | None, root: Path
    ) -> None:
        indexes = {k: v for k, v in coords.items() if isinstance(v, Index)}
        shape = tuple(map(len, indexes.values()))
        values = data.draw(float_arrays(shape=shape, allow_nan=True, allow_inf=True))
        dims = list(indexes)
        array = DataArray(values, coords, dims, name)
        save_data_array_to_disk(array, path := valid_path(root, "array"))
        view = DataArrayOnDisk(path)
        assert_identical(view.data_array, array)
        assert_identical(view.da, view.data_array)
        assert set(view.indexes) == set(indexes)
        for dim, index in view.indexes.items():
            assert_index_equal(index, indexes[dim])

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param({"x": 0}, DataArray([0, 1, 2], {"x": 0, "y": arange(3)}, ["y"])),
            param({"x": -1}, DataArray([3, 4, 5], {"x": 1, "y": arange(3)}, ["y"])),
            param(
                {"x": slice(None, 1)},
                DataArray([[0, 1, 2]], {"x": [0], "y": arange(3)}, ["x", "y"]),
            ),
            param(
                {"x": []},
                DataArray(
                    zeros((0, 3), dtype=int), {"x": [], "y": arange(3)}, ["x", "y"]
                ),
            ),
            param(
                {"x": array([True, False])},
                DataArray([[0, 1, 2]], {"x": [0], "y": arange(3)}, ["x", "y"]),
            ),
            param(
                {"x": array([0])},
                DataArray([[0, 1, 2]], {"x": [0], "y": arange(3)}, ["x", "y"]),
            ),
            param({"x": 0, "y": 0}, DataArray(0, {"x": 0, "y": 0}, [])),
            param({"x": 0, "y": -1}, DataArray(2, {"x": 0, "y": 2}, [])),
            param(
                {"x": 0, "y": slice(None, 1)}, DataArray([0], {"x": 0, "y": [0]}, ["y"])
            ),
            param({"x": 0, "y": []}, DataArray([], {"x": 0, "y": []}, ["y"])),
            param(
                {"x": 0, "y": array([True, False, False])},
                DataArray([0], {"x": 0, "y": [0]}, ["y"]),
            ),
            param({"x": 0, "y": array([0])}, DataArray([0], {"x": 0, "y": [0]}, ["y"])),
        ],
    )
    def test_isel(
        self, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        array = DataArray(
            arange(6, dtype=int).reshape(2, 3),
            {"x": arange(2), "y": arange(3)},
            ["x", "y"],
        )
        path = valid_path(tmp_path, "array")
        save_data_array_to_disk(array, path := valid_path(tmp_path, "array"))
        view = DataArrayOnDisk(path)
        assert_identical(view.isel(indexer), expected)

    @mark.parametrize(
        ("indexer", "expected"),
        [
            param(
                {"x": "x0"},
                DataArray([0, 1, 2], {"x": "x0", "y": ["y0", "y1", "y2"]}, ["y"]),
            ),
            param(
                {"x": []},
                DataArray(
                    zeros((0, 3), dtype=int),
                    {"x": [], "y": ["y0", "y1", "y2"]},
                    ["x", "y"],
                ),
            ),
            param(
                {"x": ["x0"]},
                DataArray(
                    [[0, 1, 2]], {"x": ["x0"], "y": ["y0", "y1", "y2"]}, ["x", "y"]
                ),
            ),
            param(
                {"x": ["x0", "x1"]},
                DataArray(
                    [[0, 1, 2], [3, 4, 5]],
                    {"x": ["x0", "x1"], "y": ["y0", "y1", "y2"]},
                    ["x", "y"],
                ),
            ),
            param({"x": "x0", "y": "y0"}, DataArray(0, {"x": "x0", "y": "y0"}, [])),
            param(
                {"x": "x0", "y": []},
                DataArray(zeros(0, dtype=int), {"x": "x0", "y": []}, ["y"]),
            ),
            param(
                {"x": "x0", "y": ["y0"]},
                DataArray([0], {"x": "x0", "y": ["y0"]}, ["y"]),
            ),
            param(
                {"x": "x0", "y": ["y0", "y1"]},
                DataArray([0, 1], {"x": "x0", "y": ["y0", "y1"]}, ["y"]),
            ),
        ],
    )
    def test_sel(
        self, tmp_path: Path, indexer: Mapping[str, Any], expected: Any
    ) -> None:
        array = DataArray(
            arange(6, dtype=int).reshape(2, 3),
            {"x": ["x0", "x1"], "y": ["y0", "y1", "y2"]},
            ["x", "y"],
        )
        path = valid_path(tmp_path, "array")
        save_data_array_to_disk(array, path := valid_path(tmp_path, "array"))
        view = DataArrayOnDisk(path)
        with suppress_warnings(category=FutureWarning):  # empty arrays trigger
            assert_identical(view.sel(indexer), expected)

    @mark.parametrize("func", [param(repr), param(str)])
    def test_repr_and_str(self, func: Callable[[Any], str], tmp_path: Path) -> None:
        view = DataArrayOnDisk(tmp_path)
        cls = get_class_name(DataArrayOnDisk)
        assert func(view) == f"{cls}({tmp_path})"


class TestToNDArray1:
    @given(array=int_arrays(shape=integers(0, 10)))
    def test_array(self, array: NDArrayI1) -> None:
        assert_equal(to_ndarray1(array), array)

    @given(array=int_arrays(shape=integers(0, 10)))
    def test_data_array(self, array: NDArrayI1) -> None:
        assert_equal(to_ndarray1(DataArray(array)), array)

    @given(array=int_arrays(shape=integers(0, 10)))
    def test_index(self, array: NDArrayI1) -> None:
        assert_equal(to_ndarray1(Index(array)), array)

    @mark.parametrize("array", [param(None), param(zeros(())), param(zeros((1, 1)))])
    def test_error(self, array: DataArray1) -> None:
        with raises(ToNDArray1Error):
            _ = to_ndarray1(array)


class TestYieldDataArrayOnDisk:
    @mark.parametrize("length", [param(0), param(1), param(2)])
    def test_main(self, tmp_path: Path, length: int) -> None:
        coords = {"x": RangeIndex(length)}
        with yield_data_array_on_disk(coords, valid_path(tmp_path, "array")):
            pass



from __future__ import annotations

from ast import AST, Call, ImportFrom, Module, Name, With, alias, expr, parse
from collections import Counter
from collections.abc import Iterable, Iterator
from enum import Enum, auto
from operator import itemgetter
from sys import stdout
from types import ModuleType
from typing import Any

import click
from ast_comments import Comment, unparse
from click import command, option
from loguru import logger
from typing_extensions import assert_never

from utilities.humps import snake_case
from utilities.more_itertools import one
from utilities.text import ensure_str, strip_and_dedent


class Method(Enum):
    """An enumeration of the import generation methods."""

    direct = auto()
    module = auto()
    parse = auto()


def yield_imports(
    *, method: Method = Method.module, include_suppress: bool = True
) -> Iterator[ImportFrom]:
    match method:
        case Method.direct:
            return _yield_import_nodes_directly(click, [command, option])
        case Method.module:
            import utilities.pandas

            return _yield_import_nodes_from_module_all(utilities.pandas)
        case Method.parse:
            text = """
from itertools import (
    accumulate,  # noqa: F401
    chain,  # noqa: F401
    combinations,  # noqa: F401
)
            """
            return _yield_import_nodes_from_text(
                text, include_suppress=include_suppress
            )
        case _ as never:  # type: ignore
            assert_never(never)


@command()
def main() -> None:
    imports = list(yield_imports())

    logger.info("IPython imports")
    _ = stdout.write(_generate_ipython_imports(imports) + "\n")

    logger.info("Neovim snippets")
    neovim_template = 's("{key}", {{ t("{value}") }})'
    _ = stdout.write(_generate_snippets(imports, neovim_template) + "\n")

    logger.info("VSCode snippets")
    vscode_template = '"{key}": {{ "prefix": "{key}", "body": ["{value}", "$0"] }}'
    _ = stdout.write(_generate_snippets(imports, vscode_template) + "\n")


def _yield_import_nodes_directly(
    module: ModuleType, objs: Iterable[Any], /
) -> Iterator[ImportFrom]:
    mod_name = module.__name__
    for obj in objs:
        yield ImportFrom(module=mod_name, names=[alias(name=obj.__qualname__)])


def _yield_import_nodes_from_module_all(module: ModuleType, /) -> Iterator[ImportFrom]:
    for key in module.__all__:
        yield ImportFrom(module=module.__name__, names=[alias(name=key)])


def _yield_import_nodes_from_text(
    text: str, /, *, include_suppress: bool = False
) -> Iterator[ImportFrom]:
    text = strip_and_dedent(text)
    module = parse(text)
    yield from _yield_imports_from_nodes(module.body, include_suppress=include_suppress)


def _yield_imports_from_nodes(
    nodes: Iterable[AST], /, *, include_suppress: bool = False
) -> Iterator[ImportFrom]:
    for node in nodes:
        if isinstance(node, ImportFrom):
            is_suppress = _is_suppress_import(node)
            if (is_suppress and include_suppress) or not is_suppress:
                for ali in node.names:
                    yield ImportFrom(module=node.module, names=[ali])
        elif isinstance(node, With) and _is_suppress_context_manager(
            node
        ):  # pragma: no cover
            yield from _yield_imports_from_nodes(
                node.body, include_suppress=include_suppress
            )


def _is_suppress_import(node: ImportFrom, /) -> bool:
    return (node.module == "contextlib") and any(  # pragma: no cover
        ali.name == "suppress" for ali in node.names
    )


def _is_suppress_context_manager(node: With, /) -> bool:
    return any(  # pragma: no cover
        _is_suppress_context_manager_1(i.context_expr) for i in node.items
    )


def _is_suppress_context_manager_1(node: expr, /) -> bool:
    if not isinstance(node, Call):  # pragma: no cover
        return False  # pragma: no cover
    return isinstance(func := node.func, Name) and (func.id == "suppress")


def _generate_ipython_imports(imports: Iterable[ImportFrom], /) -> str:
    def yield_imports() -> Iterator[list[AST]]:
        for imp in imports:
            yield [imp, Comment(value="# noqa: F401", inline=True)]

    body = list(yield_imports())
    mod_ast = Module(body=body, type_ignores=[])
    return unparse(mod_ast)


def _generate_snippets(imports: Iterable[ImportFrom], template: str, /) -> str:
    items = ((_node_to_key(imp), _generate_snippet(imp, template)) for imp in imports)
    sorted_items = sorted(items, key=itemgetter(0))
    counts = Counter(k for k, _ in sorted_items)
    duplicated = {k for k, v in counts.items() if v >= 2}  # noqa: PLR2004
    if len(duplicated) >= 1:
        logger.warning(f"Duplicated keys: {duplicated}")
    snippets = (f"{s}," for _, s in sorted_items)
    return "".join(snippets)


def _generate_snippet(imp: ImportFrom, template: str, /) -> str:
    key = _node_to_key(imp)
    value = unparse(imp)
    return strip_and_dedent(template).format(key=key, value=value)


def _node_to_key(imp: ImportFrom, /) -> str:
    module = ensure_str(imp.module)
    alias = one(imp.names)
    value = snake_case(alias.name).replace("_", "-")
    return f"f{module[:2]}-{value}"



from __future__ import annotations

from pathlib import Path

from pathvalidate import ValidationError, sanitize_filepath, validate_filepath

from utilities.types import PathLike


def valid_path(*parts: PathLike, sanitize: bool = False) -> Path:
    """Build & validate a path; sanitize if necessary."""
    path = Path(*parts)
    try:
        validate_filepath(path, platform="auto")
    except ValidationError:
        if sanitize:
            return sanitize_filepath(path, platform="auto")
        raise
    return path


def valid_path_cwd(*parts: PathLike, sanitize: bool = False) -> Path:
    """Build & validate a path from the current working directory."""
    return valid_path(Path.cwd(), *parts, sanitize=sanitize)


def valid_path_home(*parts: PathLike, sanitize: bool = False) -> Path:
    """Build & validate a path from home."""
    return valid_path(Path.home(), *parts, sanitize=sanitize)


__all__ = ["valid_path", "valid_path_cwd", "valid_path_home"]



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from more_itertools import always_iterable as _always_iterable
from typing_extensions import override

_T = TypeVar("_T")


def always_iterable(
    obj: _T | Iterable[_T],
    /,
    *,
    base_type: type[Any] | tuple[type[Any], ...] | None = (str, bytes),
) -> Iterator[_T]:
    """Typed version of `always_iterable`."""
    return _always_iterable(obj, base_type=base_type)


def one(iterable: Iterable[_T], /) -> _T:
    """Custom version of `one` with separate exceptions."""
    it = iter(iterable)
    try:
        first = next(it)
    except StopIteration:
        raise OneEmptyError(iterable=iterable) from None
    try:
        second = next(it)
    except StopIteration:
        return first
    raise OneNonUniqueError(iterable=iterable, first=first, second=second)


@dataclass(frozen=True, kw_only=True, slots=True)
class OneError(Exception, Generic[_T]):
    iterable: Iterable[_T]


@dataclass(frozen=True, kw_only=True, slots=True)
class OneEmptyError(OneError[_T]):
    @override
    def __str__(self) -> str:
        return "Iterable {} must contain exactly one item; it was empty".format(
            self.iterable
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class OneNonUniqueError(OneError[_T]):
    first: _T
    second: _T

    @override
    def __str__(self) -> str:
        return "Iterable {} must contain exactly one item; got {}, {} and perhaps more".format(
            self.iterable, self.first, self.second
        )


__all__ = ["OneEmptyError", "OneError", "OneNonUniqueError", "always_iterable", "one"]



from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from dataclasses import dataclass
from numbers import Number
from operator import eq, ge, gt, le, lt, ne
from timeit import default_timer
from typing import Any

from typing_extensions import Self, override

from utilities.types import EnsureClassError, ensure_class, get_class_name


class Timer:
    """Context manager for timing blocks of code."""

    def __init__(self) -> None:
        super().__init__()
        self._start = default_timer()
        self._end: float | None = None

    def __enter__(self) -> Self:
        self._start = default_timer()
        return self

    def __exit__(self, *_: object) -> bool:
        self._end = default_timer()
        return False

    def __float__(self) -> float:
        end_use = default_timer() if (end := self._end) is None else end
        return end_use - self._start

    @override
    def __repr__(self) -> str:
        return str(self.timedelta)

    @override
    def __str__(self) -> str:
        return str(self.timedelta)

    @override
    def __eq__(self, other: object) -> bool:
        return self._compare(other, eq)

    def __ge__(self, other: Any) -> bool:
        return self._compare(other, ge)

    def __gt__(self, other: Any) -> bool:
        return self._compare(other, gt)

    def __le__(self, other: Any) -> bool:
        return self._compare(other, le)

    def __lt__(self, other: Any) -> bool:
        return self._compare(other, lt)

    @override
    def __ne__(self, other: object) -> bool:
        return self._compare(other, ne)

    @property
    def timedelta(self) -> dt.timedelta:
        """The elapsed time, as a `timedelta` object."""
        return dt.timedelta(seconds=float(self))

    def _compare(self, other: Any, op: Callable[[Any, Any], bool], /) -> bool:
        try:
            right = ensure_class(other, (Number, Timer, dt.timedelta))
        except EnsureClassError:
            raise TimerError(obj=other) from None
        left = float(self) if isinstance(right, Number | Timer) else self.timedelta
        return op(left, right)


@dataclass(frozen=True, kw_only=True, slots=True)
class TimerError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return (
            "Timer must be compared to a number, Timer, or timedelta; got "
            "{} instead".format(get_class_name(self.obj))
        )


__all__ = ["Timer", "TimerError"]



from __future__ import annotations

from typing import Any, cast

import cvxpy
import numpy as np
from cvxpy import Expression, Maximize, Minimize, Problem, Variable
from numpy import array, isclose
from numpy.testing import assert_equal
from pandas import DataFrame, Series
from pandas.testing import assert_frame_equal, assert_series_equal
from pytest import mark, param, raises

from utilities.cvxpy import (
    MultiplyError,
    ScalarProductError,
    SolveInfeasibleError,
    SolveUnboundedError,
    abs_,
    add,
    divide,
    max_,
    maximum,
    min_,
    minimum,
    multiply,
    negate,
    negative,
    norm,
    positive,
    power,
    quad_form,
    scalar_product,
    solve,
    sqrt,
    subtract,
    sum_,
    sum_axis0,
    sum_axis1,
)
from utilities.numpy import NDArrayF
from utilities.pandas import SeriesF


def _get_variable(
    objective: type[Maximize] | type[Minimize],  # noqa: PYI055
    /,
    *,
    shape: tuple[int, ...] | None = None,
) -> Variable:
    if shape is None:
        var = Variable()
        scalar = var
    else:
        var = Variable(shape=shape)
        scalar = cvxpy.sum(var)
    threshold = 10.0
    problem = Problem(
        objective(scalar), [cast(Any, var) >= -threshold, cast(Any, var) <= threshold]
    )
    _ = problem.solve()
    return var


class TestAbs:
    @mark.parametrize(
        ("x", "expected"), [param(0.0, 0.0), param(1.0, 1.0), param(-1.0, 1.0)]
    )
    def test_float(self, *, x: float, expected: float) -> None:
        assert isclose(abs_(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
            param(array([-1.0]), array([1.0])),
        ],
    )
    def test_array(self, *, x: NDArrayF, expected: NDArrayF) -> None:
        assert_equal(abs_(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(Series([0.0]), Series([0.0])),
            param(Series([1.0]), Series([1.0])),
            param(Series([-1.0]), Series([1.0])),
        ],
    )
    def test_series(self, *, x: SeriesF, expected: SeriesF) -> None:
        assert_series_equal(abs_(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(DataFrame([0.0]), DataFrame([0.0])),
            param(DataFrame([1.0]), DataFrame([1.0])),
            param(DataFrame([-1.0]), DataFrame([1.0])),
        ],
    )
    def test_dataframe(self, *, x: DataFrame, expected: DataFrame) -> None:
        assert_frame_equal(abs_(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(abs_(var).value, abs_(var.value))


class TestAdd:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, 3.0),
            param(1.0, array([2.0]), array([3.0])),
            param(array([1.0]), 2.0, array([3.0])),
            param(array([1.0]), array([2.0]), array([3.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, y: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(add(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self, *, x: float | NDArrayF | Expression, objective: type[Maximize | Minimize]
    ) -> None:
        var = _get_variable(objective)
        assert isclose(add(x, var).value, add(x, var.value))  # type: ignore
        assert isclose(add(var, x).value, add(var.value, x))  # type: ignore

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(add(var1, var2).value, add(var1.value, var2.value))


class TestDivide:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, 0.5),
            param(1.0, array([2.0]), array([0.5])),
            param(array([1.0]), 2.0, array([0.5])),
            param(array([1.0]), array([2.0]), array([0.5])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, y: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(divide(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self, *, x: float | NDArrayF | Expression, objective: type[Maximize | Minimize]
    ) -> None:
        var = _get_variable(objective)
        assert_equal(divide(x, var).value, divide(x, var.value))
        assert_equal(divide(var, x).value, divide(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(divide(var1, var2).value, divide(var1.value, var2.value))


class TestMax:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(array([1.0, 2.0]), 2.0),
            param(array([-1.0, -2.0]), -1.0),
            param(Series([1.0, 2.0]), 2.0),
            param(Series([-1.0, -2.0]), -1.0),
            param(DataFrame([1.0, 2.0]), 2.0),
            param(DataFrame([-1.0, -2.0]), -1.0),
        ],
    )
    def test_float_array_and_ndframe(
        self, *, x: float | NDArrayF | SeriesF | DataFrame, expected: float
    ) -> None:
        assert_equal(max_(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert isclose(max_(var).value, max_(var.value))


class TestMaximum:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(2.0, 3.0, 3.0),
            param(2.0, array([3.0]), array([3.0])),
            param(array([2.0]), 3.0, array([3.0])),
            param(array([2.0]), array([3.0]), array([3.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, y: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(maximum(x, y), expected)

    @mark.parametrize("x", [param(2.0), param(array([2.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self, *, x: float | NDArrayF | Expression, objective: type[Maximize | Minimize]
    ) -> None:
        var = _get_variable(objective)
        assert_equal(maximum(x, var).value, maximum(x, var.value))
        assert_equal(maximum(var, x).value, maximum(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(maximum(var1, var2).value, maximum(var1.value, var2.value))


class TestMin:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(array([1.0, 2.0]), 1.0),
            param(array([-1.0, -2.0]), -2.0),
            param(Series([1.0, 2.0]), 1.0),
            param(Series([-1.0, -2.0]), -2.0),
            param(DataFrame([1.0, 2.0]), 1.0),
            param(DataFrame([-1.0, -2.0]), -2.0),
        ],
    )
    def test_float_array_and_ndframe(
        self, *, x: float | NDArrayF | SeriesF | DataFrame, expected: float
    ) -> None:
        assert isclose(min_(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective, shape=(2,))
        assert isclose(min_(var).value, min_(var.value))


class TestMinimum:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(2.0, 3.0, 2.0),
            param(2.0, array([3.0]), array([2.0])),
            param(array([2.0]), 3.0, array([2.0])),
            param(array([2.0]), array([3.0]), array([2.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, y: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(minimum(x, y), expected)

    @mark.parametrize("x", [param(2.0), param(array([2.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self, *, x: float | NDArrayF | Expression, objective: type[Maximize | Minimize]
    ) -> None:
        var = _get_variable(objective)
        assert_equal(minimum(x, var).value, minimum(x, var.value))
        assert_equal(minimum(var, x).value, minimum(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(minimum(var1, var2).value, minimum(var1.value, var2.value))


class TestMultiply:
    def test_two_floats(self) -> None:
        assert isclose(multiply(2.0, 3.0), 6.0)

    def test_two_arrays(self) -> None:
        assert_equal(multiply(array([2.0]), array([3.0])), array([6.0]))

    def test_two_series(self) -> None:
        res = multiply(Series([2.0]), Series([3.0]))
        expected = Series([6.0])
        assert_series_equal(res, expected)

    def test_two_dataframes(self) -> None:
        res = multiply(DataFrame([2.0]), DataFrame([3.0]))
        expected = DataFrame([6.0])
        assert_frame_equal(res, expected)

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert isclose(multiply(var1, var2).value, multiply(var1.value, var2.value))

    def test_float_and_array(self) -> None:
        x, y, expected = 2.0, array([3.0]), array([6.0])
        assert_equal(multiply(x, y), expected)
        assert_equal(multiply(y, x), expected)

    def test_float_and_series(self) -> None:
        x, y, expected = 2.0, Series([3.0]), Series([6.0])
        assert_series_equal(multiply(x, y), expected)
        assert_series_equal(multiply(y, x), expected)

    def test_float_and_dataframe(self) -> None:
        x, y, expected = 2.0, DataFrame([3.0]), DataFrame([6.0])
        assert_frame_equal(multiply(x, y), expected)
        assert_frame_equal(multiply(y, x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    @mark.parametrize("shape", [param(None), param((2, 2))])
    def test_float_and_expr(
        self, *, objective: type[Maximize | Minimize], shape: tuple[int, ...] | None
    ) -> None:
        x, y = 2.0, _get_variable(objective, shape=shape)
        assert_equal(multiply(x, y).value, multiply(x, y.value))
        assert_equal(multiply(y, x).value, multiply(y.value, x))

    def test_array_and_series(self) -> None:
        x, y, expected = array([2.0]), Series([3.0]), Series([6.0])
        assert_series_equal(multiply(x, y), expected)
        assert_series_equal(multiply(y, x), expected)

    def test_array_and_dataframe(self) -> None:
        x, y, expected = array([2.0]), DataFrame([3.0]), DataFrame([6.0])
        assert_frame_equal(multiply(x, y), expected)
        assert_frame_equal(multiply(y, x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_array_and_expr(self, *, objective: type[Maximize | Minimize]) -> None:
        x, y = array([2.0]), _get_variable(objective)
        assert isclose(multiply(x, y).value, multiply(x, y.value))
        assert isclose(multiply(y, x).value, multiply(y.value, x))

    def test_series_and_dataframe(self) -> None:
        x, y = Series([2.0]), DataFrame([3.0])
        with raises(MultiplyError):
            _ = multiply(cast(Any, x), cast(Any, y))
        with raises(MultiplyError):
            _ = multiply(cast(Any, y), cast(Any, x))

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_series_and_expr(self, *, objective: type[Maximize | Minimize]) -> None:
        x, y = Series([2.0]), _get_variable(objective)
        assert isclose(multiply(x, y).value, multiply(x, y.value))
        assert isclose(multiply(y, x).value, multiply(y.value, x))

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_dataframe_and_expr(self, *, objective: type[Maximize | Minimize]) -> None:
        x, y = DataFrame([2.0]), _get_variable(objective)
        assert isclose(multiply(x, y).value, multiply(x, y.value))
        assert isclose(multiply(y, x).value, multiply(y.value, x))


class TestNegate:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, -0.0),
            param(1.0, -1.0),
            param(-1.0, 1.0),
            param(array([0.0]), array([-0.0])),
            param(array([1.0]), array([-1.0])),
            param(array([-1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(negate(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(Series([0.0]), Series([0.0])),
            param(Series([1.0]), Series([-1.0])),
            param(Series([-1.0]), Series([1.0])),
        ],
    )
    def test_series(self, *, x: SeriesF, expected: SeriesF) -> None:
        assert_series_equal(negate(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(DataFrame([0.0]), DataFrame([0.0])),
            param(DataFrame([1.0]), DataFrame([-1.0])),
            param(DataFrame([-1.0]), DataFrame([1.0])),
        ],
    )
    def test_dataframe(self, *, x: DataFrame, expected: DataFrame) -> None:
        assert_frame_equal(negate(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(
        self,
        *,
        objective: type[Maximize] | type[Minimize],  # noqa: PYI055
    ) -> None:
        var = _get_variable(objective)
        assert_equal(negate(var).value, negate(var.value))


class TestNegative:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 0.0),
            param(-1.0, 1.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([0.0])),
            param(array([-1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(negative(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(
        self,
        *,
        objective: type[Maximize] | type[Minimize],  # noqa: PYI055
    ) -> None:
        var = _get_variable(objective)
        assert isclose(negative(var).value, negative(var.value))


class TestNorm:
    @mark.parametrize("x", [param(array([2.0, 3.0])), param(Series([2.0, 3.0]))])
    def test_array_and_series(self, *, x: NDArrayF | SeriesF) -> None:
        assert isclose(norm(x), np.sqrt(13))

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    @mark.parametrize("shape", [param((2,)), param((2, 2))])
    def test_expression(
        self, *, objective: type[Maximize | Minimize], shape: tuple[int, ...]
    ) -> None:
        var = _get_variable(objective, shape=shape)
        assert isclose(norm(var).value, norm(var.value))


class TestPositive:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(-1.0, 0.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
            param(array([-1.0]), array([0.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(positive(x), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(positive(var).value, positive(var.value))


class TestPower:
    @mark.parametrize(
        ("x", "p", "expected"),
        [
            param(0.0, 0.0, 1.0),
            param(2.0, 3.0, 8.0),
            param(2.0, array([3.0]), array([8.0])),
            param(array([2.0]), 3.0, array([8.0])),
            param(array([2.0]), array([3.0]), array([8.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, p: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(power(x, p), expected)

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective)
        assert_equal(power(var, 2.0).value, power(var.value, 2.0))


class TestQuadForm:
    def test_array(self) -> None:
        assert_equal(
            quad_form(array([2.0, 3.0]), array([[4.0, 5.0], [5.0, 4.0]])), 112.0
        )

    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_expression(self, *, objective: type[Maximize | Minimize]) -> None:
        var = _get_variable(objective, shape=(2,))
        P = array([[2.0, 3.0], [3.0, 2.0]])  # noqa: N806
        assert_equal(quad_form(var, P).value, quad_form(var.value, P))


class TestScalarProduct:
    @mark.parametrize(
        "x",
        [
            param(2.0),
            param(array([2.0])),
            param(Series([2.0])),
            param(DataFrame([2.0])),
        ],
    )
    @mark.parametrize(
        "y",
        [
            param(3.0),
            param(array([3.0])),
            param(Series([3.0])),
            param(DataFrame([3.0])),
        ],
    )
    def test_two_floats_arrays_and_ndframes(
        self,
        *,
        x: float | NDArrayF | SeriesF | DataFrame,
        y: float | NDArrayF | SeriesF | DataFrame,
    ) -> None:
        if (isinstance(x, Series) and isinstance(y, DataFrame)) or (
            isinstance(x, DataFrame) and isinstance(y, Series)
        ):
            with raises(ScalarProductError):
                _ = scalar_product(cast(Any, x), cast(Any, y))
            with raises(ScalarProductError):
                _ = scalar_product(cast(Any, y), cast(Any, x))
        else:
            assert isclose(scalar_product(cast(Any, y), cast(Any, x)), 6.0)
            assert isclose(scalar_product(cast(Any, x), cast(Any, y)), 6.0)

    @mark.parametrize(
        ("x", "shape"),
        [
            param(2.0, None),
            param(2.0, (2,)),
            param(2.0, (2, 2)),
            param(array([2.0]), None),
            param(array([2.0]), (1,)),
            param(array([2.0]), (2,)),
            param(array([2.0]), (1, 2)),
            param(array([2.0]), (2, 2)),
            param(Series([2.0]), None),
            param(Series([2.0]), (1,)),
            param(DataFrame([2.0]), None),
            param(DataFrame([2.0]), (1,)),
            param(DataFrame([2.0]), (1, 1)),
        ],
    )
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_float_array_and_ndframe_and_one_expression(
        self,
        *,
        x: float | NDArrayF,
        objective: type[Maximize | Minimize],
        shape: tuple[int, ...] | None,
    ) -> None:
        y = _get_variable(objective, shape=shape)
        assert isclose(scalar_product(x, y).value, scalar_product(x, y.value))
        assert isclose(scalar_product(y, x).value, scalar_product(y.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert isclose(
            scalar_product(var1, var2).value, scalar_product(var1.value, var2.value)
        )


class TestSolve:
    def test_main(self) -> None:
        var = Variable()
        problem = Problem(Minimize(sum_(abs_(var))), [])
        _ = solve(problem)

    def test_infeasible_problem(self) -> None:
        var = Variable()
        threshold = 1.0
        problem = Problem(
            Minimize(sum_(abs_(var))),
            [cast(Any, var) >= threshold, cast(Any, var) <= -threshold],
        )
        with raises(SolveInfeasibleError):
            _ = solve(problem)

    def test_unbounded_problem(self) -> None:
        var = Variable()
        problem = Problem(Maximize(sum_(var)), [])
        with raises(SolveUnboundedError):
            _ = solve(problem)


class TestSqrt:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(array([0.0]), array([0.0])),
            param(array([1.0]), array([1.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(sqrt(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [param(Series([0.0]), Series([0.0])), param(Series([1.0]), Series([1.0]))],
    )
    def test_series(self, *, x: SeriesF, expected: SeriesF) -> None:
        assert_series_equal(sqrt(x), expected)

    @mark.parametrize(
        ("x", "expected"),
        [
            param(DataFrame([0.0]), DataFrame([0.0])),
            param(DataFrame([1.0]), DataFrame([1.0])),
        ],
    )
    def test_dataframe(self, *, x: DataFrame, expected: DataFrame) -> None:
        assert_frame_equal(sqrt(x), expected)

    def test_expression(self) -> None:
        var = _get_variable(Maximize)
        assert isclose(sqrt(var).value, sqrt(var.value))


class TestSubtract:
    @mark.parametrize(
        ("x", "y", "expected"),
        [
            param(1.0, 2.0, -1.0),
            param(1.0, array([2.0]), array([-1.0])),
            param(array([1.0]), 2.0, array([-1.0])),
            param(array([1.0]), array([2.0]), array([-1.0])),
        ],
    )
    def test_float_and_array(
        self, *, x: float | NDArrayF, y: float | NDArrayF, expected: float | NDArrayF
    ) -> None:
        assert_equal(subtract(x, y), expected)

    @mark.parametrize("x", [param(1.0), param(array([1.0]))])
    @mark.parametrize("objective", [param(Maximize), param(Minimize)])
    def test_one_expression(
        self, *, x: float | NDArrayF | Expression, objective: type[Maximize | Minimize]
    ) -> None:
        var = _get_variable(objective)
        assert_equal(subtract(x, var).value, subtract(x, var.value))
        assert_equal(subtract(var, x).value, subtract(var.value, x))

    @mark.parametrize("objective1", [param(Maximize), param(Minimize)])
    @mark.parametrize("objective2", [param(Maximize), param(Minimize)])
    def test_two_expressions(
        self,
        *,
        objective1: type[Maximize | Minimize],
        objective2: type[Maximize | Minimize],
    ) -> None:
        var1 = _get_variable(objective1)
        var2 = _get_variable(objective2)
        assert_equal(subtract(var1, var2).value, subtract(var1.value, var2.value))


class TestSum:
    @mark.parametrize(
        ("x", "expected"),
        [
            param(0.0, 0.0),
            param(1.0, 1.0),
            param(-1.0, -1.0),
            param(array([0.0]), 0.0),
            param(array([1.0]), 1.0),
            param(array([-1.0]), -1.0),
            param(array([[0.0, 0.0]]), 0.0),
            param(array([[1.0, 1.0]]), 2.0),
            param(array([[-1.0, -1.0]]), -2.0),
            param(Series([0.0]), 0.0),
            param(Series([1.0]), 1.0),
            param(Series([-1.0]), -1.0),
            param(DataFrame([0.0]), 0.0),
            param(DataFrame([1.0]), 1.0),
            param(DataFrame([-1.0]), -1.0),
        ],
    )
    def test_float_array_and_ndframe(
        self, *, x: float | NDArrayF | SeriesF | DataFrame, expected: float
    ) -> None:
        assert isclose(sum_(x), expected)

    def test_expression(self) -> None:
        var = _get_variable(Maximize)
        assert_equal(sum_(var).value, sum_(var.value))


class TestSum0And1:
    def test_array(self) -> None:
        x = array([[1.0, 2.0], [3.0, 4.0]])
        assert_equal(sum_axis0(x), array([4.0, 6.0]))
        assert_equal(sum_axis1(x), array([3.0, 7.0]))

    def test_dataframe(self) -> None:
        x = DataFrame([[1.0, 2.0], [3.0, 4.0]])
        assert_series_equal(sum_axis0(x), Series([4.0, 6.0]))
        assert_series_equal(sum_axis1(x), Series([3.0, 7.0]))

    def test_expression(self) -> None:
        var = _get_variable(Maximize, shape=(2, 2))
        assert_equal(sum_axis0(var).value, sum_axis0(var.value))
        assert_equal(sum_axis1(var).value, sum_axis1(var.value))



from __future__ import annotations

import enum
from collections import defaultdict
from collections.abc import Iterable, Iterator
from contextlib import suppress
from dataclasses import dataclass
from enum import auto
from itertools import chain
from math import floor
from typing import Any, TypeGuard, cast

import sqlalchemy
from more_itertools import chunked
from sqlalchemy import Column, Connection, Engine, Table, insert
from sqlalchemy.dialects.mssql import dialect as mssql_dialect
from sqlalchemy.dialects.mysql import dialect as mysql_dialect
from sqlalchemy.dialects.oracle import dialect as oracle_dialect
from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.exc import ArgumentError, DatabaseError
from sqlalchemy.orm import InstrumentedAttribute, class_mapper
from sqlalchemy.orm.exc import UnmappedClassError
from typing_extensions import assert_never, override

from utilities.errors import redirect_error
from utilities.iterables import is_iterable_not_str
from utilities.more_itertools import one
from utilities.types import get_class_name

CHUNK_SIZE_FRAC = 0.95


class Dialect(enum.Enum):
    """An enumeration of the SQL dialects."""

    mssql = auto()
    mysql = auto()
    oracle = auto()
    postgresql = auto()
    sqlite = auto()

    @property
    def max_params(self, /) -> int:
        match self:
            case Dialect.mssql:  # pragma: no cover
                return 2100
            case Dialect.mysql:  # pragma: no cover
                return 65535
            case Dialect.oracle:  # pragma: no cover
                return 1000
            case Dialect.postgresql:  # pragma: no cover
                return 32767
            case Dialect.sqlite:
                return 100
            case _ as never:  # type: ignore
                assert_never(never)


def ensure_tables_created(
    engine: Engine, /, *tables_or_mapped_classes: Table | type[Any]
) -> None:
    """Ensure a table/set of tables is/are created."""

    class TableAlreadyExistsError(Exception):
        ...

    match dialect := get_dialect(engine):
        case Dialect.mysql | Dialect.postgresql:  # pragma: no cover
            raise NotImplementedError(dialect)
        case Dialect.mssql:  # pragma: no cover
            match = "There is already an object named .* in the database"
        case Dialect.oracle:  # pragma: no cover
            match = "ORA-00955: name is already used by an existing object"
        case Dialect.sqlite:
            match = "table .* already exists"
        case _ as never:  # type: ignore
            assert_never(never)

    for table_or_mapped_class in tables_or_mapped_classes:
        table = get_table(table_or_mapped_class)
        with suppress(TableAlreadyExistsError), redirect_error(
            DatabaseError, TableAlreadyExistsError, match=match
        ), engine.begin() as conn:
            table.create(conn)


def get_chunk_size(
    engine_or_conn: Engine | Connection,
    /,
    *,
    chunk_size_frac: float = CHUNK_SIZE_FRAC,
    scaling: float = 1.0,
) -> int:
    """Get the maximum chunk size for an engine."""

    dialect = get_dialect(engine_or_conn)
    max_params = dialect.max_params
    return max(floor(chunk_size_frac * max_params / scaling), 1)


def get_column_names(table_or_mapped_class: Table | type[Any], /) -> list[str]:
    """Get the column names from a table or model."""
    return [col.name for col in get_columns(table_or_mapped_class)]


def get_columns(table_or_mapped_class: Table | type[Any], /) -> list[Column[Any]]:
    """Get the columns from a table or model."""
    return list(get_table(table_or_mapped_class).columns)


def get_dialect(engine_or_conn: Engine | Connection, /) -> Dialect:
    """Get the dialect of a database."""
    dialect = engine_or_conn.dialect
    if isinstance(dialect, mssql_dialect):  # pragma: os-ne-linux
        return Dialect.mssql
    if isinstance(dialect, mysql_dialect):  # pragma: os-ne-linux
        return Dialect.mysql
    if isinstance(dialect, oracle_dialect):
        return Dialect.oracle
    if isinstance(dialect, postgresql_dialect):  # pragma: os-ne-linux
        return Dialect.postgresql
    if isinstance(dialect, sqlite_dialect):
        return Dialect.sqlite
    raise GetDialectError(dialect=dialect)  # pragma: no cover


@dataclass(frozen=True, kw_only=True, slots=True)
class GetDialectError(Exception):
    dialect: sqlalchemy.Dialect

    @override
    def __str__(self) -> str:
        return (  # pragma: no cover
            "Dialect must be one of MS SQL, MySQL, Oracle, PostgreSQL or SQLite; got {} instead".format(
                self.dialect
            )
        )


def get_table(obj: Table | type[Any], /) -> Table:
    """Get the table from a Table or mapped class."""
    if isinstance(obj, Table):
        return obj
    if is_mapped_class(obj):
        return cast(Any, obj).__table__
    raise GetTableError(obj=obj)


@dataclass(frozen=True, kw_only=True, slots=True)
class GetTableError(Exception):
    obj: Any

    @override
    def __str__(self) -> str:
        return "Object {} must be a Table or mapped class; got {!r}".format(
            self.obj, get_class_name(self.obj)
        )


def insert_items(
    engine: Engine, *items: Any, chunk_size_frac: float = CHUNK_SIZE_FRAC
) -> None:
    """Insert a set of items into a database.

    These can be either a:
     - tuple[Any, ...], table
     - dict[str, Any], table
     - [tuple[Any ,...]], table
     - [dict[str, Any], table
     - Model
    """

    dialect = get_dialect(engine)
    to_insert: dict[Table, list[_InsertItemValues]] = defaultdict(list)
    lengths: set[int] = set()
    for item in chain(*map(_insert_items_collect, items)):
        values = item.values
        to_insert[item.table].append(values)
        lengths.add(len(values))
    max_length = max(lengths, default=1)
    chunk_size = get_chunk_size(
        engine, chunk_size_frac=chunk_size_frac, scaling=max_length
    )
    for table, values in to_insert.items():
        ensure_tables_created(engine, table)
        ins = insert(table)
        with engine.begin() as conn:
            for chunk in chunked(values, n=chunk_size):
                if dialect is Dialect.oracle:  # pragma: no cover
                    _ = conn.execute(ins, cast(Any, chunk))
                else:
                    _ = conn.execute(ins.values(list(chunk)))


_InsertItemValues = tuple[Any, ...] | dict[str, Any]


@dataclass
class _InsertionItem:
    values: _InsertItemValues
    table: Table


def _insert_items_collect(item: Any, /) -> Iterator[_InsertionItem]:
    """Collect the insertion items."""
    if isinstance(item, tuple):
        with redirect_error(ValueError, _InsertItemsCollectError(f"{item=}")):
            data, table_or_mapped_class = item
        if not is_table_or_mapped_class(table_or_mapped_class):
            msg = f"{table_or_mapped_class=}"
            raise _InsertItemsCollectError(msg)
        if _insert_items_collect_valid(data):
            yield _InsertionItem(values=data, table=get_table(table_or_mapped_class))
        elif is_iterable_not_str(data):
            yield from _insert_items_collect_iterable(data, table_or_mapped_class)
        else:
            msg = f"{data=}"
            raise _InsertItemsCollectError(msg)
    elif is_iterable_not_str(item):
        for i in item:
            yield from _insert_items_collect(i)
    elif is_mapped_class(cls := type(item)):
        yield _InsertionItem(values=mapped_class_to_dict(item), table=get_table(cls))
    else:
        msg = f"{item=}"
        raise _InsertItemsCollectError(msg)


class _InsertItemsCollectError(Exception):
    ...


def _insert_items_collect_iterable(
    obj: Iterable[Any], table_or_mapped_class: Table | type[Any], /
) -> Iterator[_InsertionItem]:
    """Collect the insertion items, for an iterable."""
    table = get_table(table_or_mapped_class)
    for datum in obj:
        if _insert_items_collect_valid(datum):
            yield _InsertionItem(values=datum, table=table)
        else:
            msg = f"{datum=}"
            raise _InsertItemsCollectIterableError(msg)


class _InsertItemsCollectIterableError(Exception):
    ...


def _insert_items_collect_valid(obj: Any, /) -> TypeGuard[_InsertItemValues]:
    """Check if an insertion item being collected is valid."""
    return isinstance(obj, tuple) or (
        isinstance(obj, dict) and all(isinstance(key, str) for key in obj)
    )


def is_mapped_class(obj: type[Any], /) -> bool:
    """Check if an object is a mapped class."""

    try:
        _ = class_mapper(obj)
    except (ArgumentError, UnmappedClassError):
        return False
    return True


def is_table_or_mapped_class(obj: Table | type[Any], /) -> bool:
    """Check if an object is a Table or a mapped class."""

    return isinstance(obj, Table) or is_mapped_class(obj)


def mapped_class_to_dict(obj: Any, /) -> dict[str, Any]:
    """Construct a dictionary of elements for insertion."""
    cls = type(obj)

    def is_attr(attr: str, key: str, /) -> str | None:
        if isinstance(value := getattr(cls, attr), InstrumentedAttribute) and (
            value.name == key
        ):
            return attr
        return None

    def yield_items() -> Iterator[tuple[str, Any]]:
        for key in get_column_names(cls):
            attr = one(attr for attr in dir(cls) if is_attr(attr, key) is not None)
            yield key, getattr(obj, attr)

    return dict(yield_items())


__all__ = [
    "CHUNK_SIZE_FRAC",
    "Dialect",
    "GetDialectError",
    "GetTableError",
    "ensure_tables_created",
    "get_chunk_size",
    "get_column_names",
    "get_columns",
    "get_dialect",
    "get_table",
    "insert_items",
    "is_mapped_class",
    "is_table_or_mapped_class",
    "mapped_class_to_dict",
]



from __future__ import annotations

from semver import Version


def ensure_version(version: Version | str, /) -> Version:
    """Ensure the object is a `Version`."""
    return version if isinstance(version, Version) else Version.parse(version)


__all__ = ["ensure_version"]



from __future__ import annotations

from typing import Any, cast

from utilities.airium import yield_airium
from utilities.text import strip_and_dedent


class TestYieldAirium:
    def test_main(self) -> None:
        with yield_airium() as airium, cast(Any, airium).div():
            airium("hello")
        result = str(airium)
        expected = """
            <!DOCTYPE html>
            <html>
              <body>
                <div>
                  hello
                </div>
              </body>
            </html>
        """
        assert result == strip_and_dedent(expected)



from __future__ import annotations

from typing import Any

from hypothesis import given
from hypothesis.strategies import DataObject, data
from luigi import Task

from utilities.hypothesis import namespace_mixins


class TestNamespaceMixins:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        _ = data.draw(namespace_mixins())

    @given(namespace_mixin=namespace_mixins())
    def test_first(self, *, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        _ = Example()

    @given(namespace_mixin=namespace_mixins())
    def test_second(self, *, namespace_mixin: Any) -> None:
        class Example(namespace_mixin, Task):
            ...

        _ = Example()



from __future__ import annotations

import datetime as dt
from collections.abc import Hashable
from typing import Any

from hypothesis import assume
from hypothesis.extra.pandas import indexes as _indexes
from hypothesis.strategies import (
    DrawFn,
    SearchStrategy,
    composite,
    dates,
    datetimes,
    integers,
)
from numpy import int64
from pandas import Timedelta, Timestamp

from utilities._hypothesis.common import MaybeSearchStrategy, lift_draw, text_ascii
from utilities._hypothesis.numpy import int64s
from utilities.datetime import UTC
from utilities.pandas import (
    TIMESTAMP_MAX_AS_DATE,
    TIMESTAMP_MAX_AS_DATETIME,
    TIMESTAMP_MIN_AS_DATE,
    TIMESTAMP_MIN_AS_DATETIME,
    IndexA,
    IndexI,
    IndexS,
    rename_index,
    sort_index,
    string,
)


@composite
def dates_pd(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.date] = TIMESTAMP_MIN_AS_DATE,
    max_value: MaybeSearchStrategy[dt.date] = TIMESTAMP_MAX_AS_DATE,
) -> dt.date:
    """Strategy for generating dates which can become Timestamps."""
    draw = lift_draw(_draw)
    return draw(dates(min_value=draw(min_value), max_value=draw(max_value)))


@composite
def datetimes_pd(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MIN_AS_DATETIME,
    max_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MAX_AS_DATETIME,
) -> dt.datetime:
    """Strategy for generating datetimes which can become Timestamps."""
    draw = lift_draw(_draw)
    datetime = draw(
        datetimes(
            min_value=draw(min_value).replace(tzinfo=None),
            max_value=draw(max_value).replace(tzinfo=None),
        )
    )
    return datetime.replace(tzinfo=UTC)


_INDEX_LENGTHS = integers(0, 10)


@composite
def indexes(
    _draw: DrawFn,
    /,
    *,
    elements: SearchStrategy[Any] | None = None,
    dtype: Any = None,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> IndexA:
    """Strategy for generating Indexes."""
    draw = lift_draw(_draw)
    n_ = draw(n)
    index = draw(
        _indexes(
            elements=elements,
            dtype=dtype,
            min_size=n_,
            max_size=n_,
            unique=draw(unique),
        )
    )
    index = rename_index(index, draw(name))
    if draw(sort):
        return sort_index(index)
    return index


def int_indexes(
    *,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> SearchStrategy[IndexI]:
    """Strategy for generating integer Indexes."""
    return indexes(
        elements=int64s(), dtype=int64, n=n, unique=unique, name=name, sort=sort
    )


@composite
def str_indexes(
    _draw: DrawFn,
    /,
    *,
    min_size: MaybeSearchStrategy[int] = 0,
    max_size: MaybeSearchStrategy[int | None] = None,
    n: MaybeSearchStrategy[int] = _INDEX_LENGTHS,
    unique: MaybeSearchStrategy[bool] = True,
    name: MaybeSearchStrategy[Hashable] = None,
    sort: MaybeSearchStrategy[bool] = False,
) -> IndexS:
    """Strategy for generating string Indexes."""
    draw = lift_draw(_draw)
    elements = text_ascii(min_size=min_size, max_size=max_size)
    index = draw(
        indexes(
            elements=elements, dtype=object, n=n, unique=unique, name=name, sort=sort
        )
    )
    return index.astype(string)


@composite
def timestamps(
    _draw: DrawFn,
    /,
    *,
    min_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MIN_AS_DATETIME,
    max_value: MaybeSearchStrategy[dt.datetime] = TIMESTAMP_MAX_AS_DATETIME,
    allow_nanoseconds: MaybeSearchStrategy[bool] = False,
) -> Timestamp:
    """Strategy for generating Timestamps."""
    draw = lift_draw(_draw)
    min_value, max_value = (draw(mv) for mv in (min_value, max_value))
    datetime = draw(datetimes_pd(min_value=min_value, max_value=max_value))
    timestamp: Timestamp = Timestamp(datetime)
    if draw(allow_nanoseconds):
        nanoseconds = draw(integers(-999, 999))
        timedelta = Timedelta(nanoseconds=nanoseconds)  # type: ignore
        timestamp += timedelta
        _ = assume(min_value <= timestamp.floor("us"))
        _ = assume(timestamp.ceil("us") <= max_value)
    return timestamp


__all__ = [
    "dates_pd",
    "datetimes_pd",
    "indexes",
    "int_indexes",
    "str_indexes",
    "timestamps",
]



from __future__ import annotations

from click import command, echo
from click.testing import CliRunner
from hypothesis import given
from hypothesis.strategies import integers, none
from pytest import mark, param

from utilities.click import (
    local_scheduler_option_default_central,
    local_scheduler_option_default_local,
    workers_option,
)
from utilities.types import SequenceStrs


class TestLocalSchedulerOption:
    @mark.parametrize(
        ("args", "expected"),
        [param([], True), param(["-ls"], True), param(["-nls"], False)],
    )
    def test_default_local(self, *, args: SequenceStrs, expected: bool) -> None:
        @command()
        @local_scheduler_option_default_local
        def cli(*, local_scheduler: bool) -> None:
            echo(f"local_scheduler = {local_scheduler}")

        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"local_scheduler = {expected}\n"

    @mark.parametrize(
        ("args", "expected"),
        [param([], False), param(["-ls"], True), param(["-nls"], False)],
    )
    def test_default_central(self, *, args: SequenceStrs, expected: bool) -> None:
        @command()
        @local_scheduler_option_default_central
        def cli(*, local_scheduler: bool) -> None:
            echo(f"local_scheduler = {local_scheduler}")

        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"local_scheduler = {expected}\n"


class TestWorkersOption:
    @given(workers=integers() | none())
    def test_main(self, workers: int | None) -> None:
        @command()
        @workers_option
        def cli(*, workers: int | None) -> None:
            echo(f"workers = {workers}")

        args = [] if workers is None else ["--workers", str(workers)]
        result = CliRunner().invoke(cli, args)
        assert result.exit_code == 0
        assert result.stdout == f"workers = {workers}\n"
